function addCardListObserver() {
debug('add observer for cards/lists');
	dirObserver.setUp();
}

function removeCardListObserver() {
debug('remove observer for cards/lists');
	dirObserver.cleanUp();
}

/*
	lastModifiedDate of a directory is only used by TB with ldap
	but i want it for normal directories
*/
var dirObserver={
  topics: [
    "addrbook-directory-created",		//nsIAbDirectory, null
    "addrbook-directory-updated",		//nsIAbDirectory, DirName //after change of name
                                    //TODO: probably should show a warning if addrbook was synchronized
    "addrbook-directory-deleted",		//nsIAbDirectory, null
    "addrbook-contact-created",			//nsIAbCard, book.UID (==itemAdded)
    "addrbook-contact-updated",			//nsIAbCard, book.UID (==itemPropertyChanged with nsIAbCard)
    "addrbook-contact-properties-updated",  //nsIAbCard, JSON.stringify(data) //see AddrBookDirectory.jsm
    "addrbook-contact-deleted",			//nsIAbCard, book.UID (==itemRemoved)
    "addrbook-list-created",				//nsIAbDirectory, book.UID
    "addrbook-list-updated",				//nsIAbDirectory, book.UID
    "addrbook-list-deleted",				//nsIAbDirectory, book.UID
    "addrbook-list-member-added",		//nsIAbCard, list.UID
    "addrbook-list-member-removed",	//nsIAbCard, list.UID
  ],
//AddrBookManager.getDirectoryFromUID(uid)
  setUp() {
    for (let topic of this.topics) {
try {
      Services.obs.addObserver(dirObserver, topic);
} catch(e) { debug('listener setUp for topic "'+topic+'" '+(typeof dirObserver)+' throws: '+e, ''); }
    }
  },
  cleanUp() {
    for (let topic of this.topics) {
try {
      Services.obs.removeObserver(dirObserver, topic);
//sometimes throws NS_ERROR_ILLEGAL_VALUE (Mail from Albrecht Dreß, 14.11.20)
//unknown topic throws NS_ERROR_FAILURE
//null topic or null observer throws NS_ERROR_ILLEGAL_VALUE
//Seems that the problem comes from a bad profile (according to Albrechts)
} catch(e) { debug('listener cleanUp for topic "'+topic+'" '+(typeof dirObserver)+' throws: '+e, ''); }
    }
  },
	bookUID: null,
	listUID: null,
	card: null,
	list: null,
	dir: null,
  props: null,

	observe(subject, topic, data) {
    //this.events.push([topic, subject, data]);
debug(topic+' '+subject+' '+data);
    switch (topic) {
			case "addrbook-contact-created":
			case "addrbook-contact-deleted":
				bookUID=data;
				card=subject;
				dir=MailServices.ab.getDirectoryFromUID(bookUID);
debug('		'+topic+': parent='+dir.dirName+' email='+card.primaryEmail+' name='+card.lastName);
        sync(dir);
				break;
			case "addrbook-contact-updated":
				bookUID=data;
				card=subject;
				dir=MailServices.ab.getDirectoryFromUID(bookUID);
        //check for sync is done in "addrbook-contact-properties-updated"
debug('		'+topic+': parent='+dir.dirName+' email='+card.primaryEmail+' name='+card.lastName);
        break;
			case "addrbook-contact-properties-updated":
				card=subject;
				dir=MailServices.ab.getDirectoryFromUID(card.directoryUID);
        props=JSON.parse(data);
        delete props.PopularityIndex; //This is the only change if history is used
debug('		'+topic+': parent='+dir.dirName+' email='+card.primaryEmail+' changed properties='+JSON.stringify(props));
        if (Object.keys(props).length > 0)
           sync(dir);
        break;
			case "addrbook-list-created":
			case "addrbook-list-updated":
			case "addrbook-list-deleted":
				bookUID=data;
				list=subject;
				dir=MailServices.ab.getDirectoryFromUID(bookUID);
debug('		'+topic+': parent='+dir.dirName+' list='+list.dirName+' '+list.UID);
        sync(dir);
				break;
			case "addrbook-list-member-added":
			case "addrbook-list-member-removed":
				listUID=data;
				card=subject;
				//list=MailServices.ab.getDirectoryFromUID(listUID); //returns null
        if (card.directoryUID) {  //since TB8x
          dir=MailServices.ab.getDirectoryFromUID(card.directoryUID);
        } else {  //TB78  card.directoryId is something like "ldap_2.servers.prefname&dirname"
          let dirId=card.directoryId;	//something like ldap_2.servers.prefname&dirname
debug('		TB78: directoryId='+dirId);
          dir=MailServices.ab.getDirectoryFromId(dirId.replace(/\&[^\&]*$/,''));
        }
debug('		'+topic+': parent='+dir.dirName);
        sync(dir);
				break;
      case "addrbook-directory-updated":		//nsIAbDirectory, "DirName" //after change of name
        dir=subject;
debug('		'+topic+': change of '+data+' new name='+dir.dirName);
        //nothing to do
        storage();
        break;
      case "addrbook-directory-created":		//nsIAbDirectory, null
        dir=subject;
debug('		'+topic+': created='+dir.dirName);
        if (dir.dirType==kJSDirectory  //only .mab/.sqlite books (exclude carddav)
            && Services.prefs.getCharPref(dir.dirPrefId+'.tbSyncAccountID', '')=="") // && not tbsync
        {
          let pref=dir.dirPrefId;
          u2i[dir.UID]=pref; //needed by addressbooks window
          prefs[pref+'.up']=false;
          prefs[pref+'.down']=false;
        }
        break;
      case "addrbook-directory-deleted":		//nsIAbDirectory, null
        dir=subject;
debug('		'+topic+': removed='+dir.dirName);
        delete u2i[dir.UID]; //needed by addressbooks window
        let pref=dir.dirPrefId;
        delete prefs[pref+'.up'];
        delete prefs[pref+'.down'];
        delete prefs[pref+'.filename'];
        delete autoupload[pref]; //just in case
        storage();
        break;
		}
  }
}

function sync(dir) {
  let now=Math.floor(Date.now()/1000);
debug('     lmd of dir set from '+dir.lastModifiedDate+' to '+now);
  dir.lastModifiedDate=now;
  let pref=dir.dirPrefId;
  if (prefs.autoupload && prefs[pref+'.up']) {
debug('       start upload timer');
    autoupload[pref]=dir;
    startUploadTimer();
  }
}
async function storage() {
}